/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met: 
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <stdlib.h>
#include <stdio.h>

#include "table.h"
#include "table_private.h"

table_t *table_create(table_cmp_func *cmpfunc){
    table_t *table;
    table = calloc(sizeof(table_t), 1);
    table->cmpfunc = cmpfunc;
	
    return table;
}

/* single rotation, for splay tree */
static void rotate(table_t *t, table_node_t *s, table_node_t *parent){
    table_node_t *temp;
	
    if(parent->left == s){
        temp = s->right;
		
        s->right = parent;
        parent->parent = s;
		
        parent->left = temp;
        if(temp != NULL)
            temp->parent = parent;
    }else{
        temp = s->left;
		
        s->left = parent;
        parent->parent = s;
		
        parent->right = temp;
        if(temp != NULL)
            temp->parent = parent;
    }
	
    t->root = s;
    s->parent = NULL;
}

/* one kind of double rotation, for splaying */
static void zigzag(table_t *t, table_node_t *s, table_node_t *parent, table_node_t *gparent){
    table_node_t *ggparent;
    table_node_t *temp1, *temp2;
	
    ggparent = gparent->parent;   /* may be NULL, if gparent was root */
    temp1 = s->left;
    temp2 = s->right;
	
    if(parent == gparent->left){
        /* left hand style zig-zag.  s == parent->right */
		
        /* move s to top */
        s->left = parent;
        parent->parent = s;
		
        s->right = gparent;
        gparent->parent = s;
		
        /* reattach subtrees */
        parent->right = temp1;
        if(temp1 != NULL)
            temp1->parent = parent;
		
        gparent->left = temp2;
        if(temp2 != NULL)
            temp2->parent = gparent;
    }else{
        /* right hand style zig-zag.  s == parent->left */
		
        /* move s to top */
        s->right = parent;
        parent->parent = s;
		
        s->left = gparent;
        gparent->parent = s;
		
        /* reattach subtrees */
        gparent->right = temp1;
        if(temp1 != NULL)
            temp1->parent = gparent;
		
        parent->left = temp2;
        if(temp2 != NULL)
            temp2->parent = parent;
    }
	
    /* now fix ggparent and s->parent */
    s->parent = ggparent;
    if(ggparent == NULL){
        /* s is suddenly the root */
        t->root = s;
    }else{
        if(ggparent->left == gparent){
            ggparent->left = s;
        }else{
            ggparent->right = s;
        }
    }
}

/* zigzig rotation for splaying */
static void zigzig(table_t *t, table_node_t *s, table_node_t *parent, table_node_t *gparent){
    table_node_t *ggparent;
    table_node_t *temp1, *temp2;
	
    ggparent = gparent->parent;   /* may be NULL, if gparent was root */
	
    if(parent->left == s){
        /* left hand zig-zig */
        
        /* tempX goes from smallest to largest valued subtree */
        temp1 = s->right;
        temp2 = parent->right;
		
        s->right = parent;
        parent->parent = s;
		
        parent->right = gparent;
        gparent->parent = parent;
		
        /* reattach lost subtrees */
        parent->left = temp1;
        if(temp1 != NULL)
            temp1->parent = parent;
		
        gparent->left = temp2;
        if(temp2 != NULL)
            temp2->parent = gparent;
    }else{
        /* right hand zig-zig */
        temp1 = parent->left;
        temp2 = s->left;
		
        s->left = parent;
        parent->parent = s;
		
        parent->left = gparent;
        gparent->parent = parent;
		
        /* reattach lost subtrees */
        gparent->right = temp1;
        if(temp1 != NULL)
            temp1->parent = gparent;
		
        parent->right = temp2;
        if(temp2 != NULL)
            temp2->parent = parent;
    }
	
    /* now fix ggparent and s->parent */
    s->parent = ggparent;
    if(ggparent == NULL){
        /* s is suddenly the root */
        t->root = s;
    }else{
        if(ggparent->left == gparent){
            ggparent->left = s;
        }else{
            ggparent->right = s;
        }
    }
}

/* splay tree splaying, calls rotate functions until s is root */
static void splay(table_t *t, table_node_t *s){
    table_node_t *parent, *gparent;
    while(t->root != s){
        parent = s->parent;
        gparent = parent->parent;
        if(gparent == NULL){
            /* s is a single rotate away from the root */
            rotate(t, s, parent);
        }else if((parent->left == s && gparent->left == parent)
			||(parent->right == s && gparent->right == parent)){
            zigzig(t, s, parent, gparent);
        }else{
            zigzag(t, s, parent, gparent);
        }
    }
}

static error_t _table_destroy(table_node_t *t, table_free_func *freefunc){
    error_t ret;
    
    if(t == NULL){
        return 0;
    }
	
    ret = _table_destroy(t->left, freefunc) || _table_destroy(t->right, freefunc);
    if(freefunc != NULL){
        freefunc(t->data);
    }
    free(t);
    return ret;
}

error_t table_destroy(table_t *t, table_free_func *freefunc){
    error_t ret;
    
    if(t == NULL){
        printf("Invalid tree!\n");
        return -1;
    }
	
    ret = _table_destroy(t->root, freefunc);
    free(t);
    return ret;
}

error_t table_store(table_t *t, void *data){
    int compare, done = 0;
    table_node_t *s = t->root;
    
    if(t == NULL){
        printf("Invalid tree!\n");
        return -1;
    }
	
    if(t->root == NULL){
        t->root = calloc(sizeof(table_node_t), 1);
        t->root->data = data;
        return 0;
    }
    
    while(!done){
        compare = t->cmpfunc(data, s->data);
        if(compare < 0){
            /* go left */
            if(s->left == NULL){
                s->left = calloc(sizeof(*(s->left)), 1);
                s->left->data = data;
                s->left->parent = s;
                s = s->left;
                done = 1;
            }else{
                s = s->left;
            }
        }else{
            /* go right */
            if(s->right == NULL){
                s->right = calloc(sizeof(*(s->right)), 1);
                s->right->data = data;
                s->right->parent = s;
                s = s->right;
                done = 1;
            }else{
                s = s->right;
            }
        }
    }
	
    /* splay the tree */
    splay(t, s);
	
    return 0;
}

error_t table_delete(table_t *t, void *data, table_free_func *freefunc){
    int compare;
    table_node_t **ppath = &(t->root); /* path taken from parent */
    table_node_t *c = t->root, *parent = NULL;
	
    if(t == NULL){
        printf("Invalid tree!\n");
        return -1;
    }
	
    if(t->root == NULL){
        return -1;
    }
	
    while(c != NULL){
        compare = t->cmpfunc(data, c->data);
        if(compare == 0){
            /* we found it */
            if(freefunc != NULL){
                freefunc(c->data);
            }
			
            if(c->left == NULL && c->right == NULL){
                /* we are a leaf */
                *ppath = NULL;
                free(c);
            }else if(c->left == NULL){
                /* we have one child, a right */
                *ppath = c->right;
                c->right->parent = c->parent;
                free(c);
            }else if(c->right == NULL){
                /* we have one child, a left */
                *ppath = c->left;
                c->left->parent = c->parent;
                free(c);
            }else{
                /* we have two children */
				
                /* save left subtree of c */
                table_node_t *left = c->left;
                
                /* put right subtree in c's place */
                *ppath = c->right;
                c->right->parent = c->parent;
                free(c);
                c = *ppath;
                
                /* subtree pointed to be left is smaller than anything
				* under the new c (the right subtree) */
                while(c->left != NULL){
                    c = c->left;
                }
                c->left = left;
                left->parent = c;
            }
            
            /* splay the tree (unless at the root) */
            if(parent != NULL){
                splay(t, parent);
            }
            
            return 0;
        }else if(compare < 0){
            /* go left */
            ppath = &(c->left);
            parent = c;
            c = c->left;
        }else{
            /* go right */
            ppath = &(c->right);
            parent = c;
            c = c->right;
        }
    }
    
    /* it wasn't in the tree */
    return -1;
}

void *table_find(table_t *t, void *data){
    table_node_t *c = t->root, *p = NULL;
    int compare;
	
    while(c != NULL){
        p = c;
        compare = t->cmpfunc(data, c->data);
        if(compare == 0){
            /* found it */
            splay(t, c);
            return c->data;
        }else if(compare < 0){
            c = c->left;
        }else{
            c = c->right;
        }
    }
    /* never found it */
    if(p != NULL){
        splay(t, p);
    }
	
    return NULL;
	
}

/* note: inorder traversal. */
static void table_apply_rec(table_node_t *curr, table_apply_func *applyfunc,
							void *extra)
{
    if ( curr == NULL )
        return;
	
    table_apply_rec( curr->left, applyfunc, extra );
    applyfunc( curr->data, extra );
    table_apply_rec( curr->right, applyfunc, extra );
}

void table_foreach(table_t *t, table_apply_func *applyfunc, void *extra)
{
    if ( t == NULL )
        return;
	
    table_apply_rec( t->root, applyfunc, extra );
}

